#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 18/03/2020
           """
__all__ = ["default_datetime_repr"]

import datetime

DEFAULT_REPRESENTATION = "%Y-%m-%d_%H:%M:%S.%f"


def default_datetime_repr(date: datetime, str_format: str = DEFAULT_REPRESENTATION) -> str:
    """

    :param date:
    :type date:
    :param str_format:
    :type str_format:
    :return:
    :rtype:"""
    return date.strftime(str_format)


def now_repr() -> str:
    """ """
    return default_datetime_repr(datetime.datetime.now())


if __name__ == "__main__":
    print(now_repr())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 26-01-2021
           """

__all__ = ["FirstArgIdentifier", "get_first_arg_name", "cprint"]

import ast
from typing import Optional, Any


def recurse_first_args(args):
    for a in args:
        if isinstance(a, ast.Call):
            if isinstance(a.func, ast.Attribute):
                r = f"{a.func.attr}"
            else:
                r = f"{a.func.id}"
            return f"{r}({recurse_first_args(a.args)})"
    else:
        if isinstance(args, ast.Constant):
            return args.value


class FirstArgIdentifier(ast.NodeVisitor):
    """ """

    def __init__(
        self,
        *args,
        verbose: bool = False,
        max_num_intermediate_unnamed_elements: int = -1,  # recurse = -1
    ):
        if len(args) < 1:
            raise ValueError("Supply at least one target function")
        self.result = {arg: {} for arg in args}
        self.verbose = verbose
        # assert max_num_intermediate_unnamed_elements >= 0
        self.sequence_depth = max_num_intermediate_unnamed_elements

    def visit_Call(self, node: ast.AST) -> None:
        """
        Should work for most use cases, but no guarantee

        :param node:
        :return:
        """
        if hasattr(node.func, "id") and node.func.id in self.result:
            first_arg = node.args[0]
            if isinstance(first_arg, ast.Name):
                iter_name = first_arg.id
            elif isinstance(first_arg, ast.Call):
                if isinstance(first_arg.func, ast.Attribute):
                    # print(first_arg.func.value.value.id) # TODO: UNROLLING is possible, do some resursion
                    # print(first_arg.func.value.attr) # TODO: SAME for full qualification in scope
                    iter_name = f"{first_arg.func.attr}"
                else:
                    iter_name = f"{first_arg.func.id}"

                if self.sequence_depth < 0:
                    iter_name += f"({recurse_first_args(first_arg.args)})"

                if self.verbose:
                    args_repr = f'{", ".join([ast.dump(sub) for sub in first_arg.args])}'
                    kws_repr = f'{", ".join([ast.dump(sub) for sub in first_arg.keywords])}'
                    args_kw_repr = []
                    if len(args_repr) > 1:
                        args_kw_repr.append(args_repr)
                    if len(kws_repr) > 1:
                        args_kw_repr.append(kws_repr)
                    iter_name += f'({", ".join(args_kw_repr)})'
            elif isinstance(first_arg, (ast.List, ast.Set, ast.Tuple)):
                elts = first_arg.elts
                if self.sequence_depth > 0:
                    if self.sequence_depth < len(elts) - 2:
                        if (
                            self.sequence_depth
                        ):  # TODO: Generalise to another external function, "pick num from sequence" func
                            stride = (len(elts) // self.sequence_depth) + 1
                            between = elts[1:-2:stride]
                        else:
                            between = []
                        elts_str = [ast.dump(sub) for sub in [elts[0]] + between + [elts[-1]]]
                        iter_name = f'[{" .. ".join(elts_str)}]'
                    else:
                        iter_name = f'[{", ".join([ast.dump(sub) for sub in elts])}]'
                else:
                    iter_name = f'[{", ".join([ast.dump(sub) for sub in elts])}]'
            elif isinstance(first_arg, ast.Dict):
                kw_repr = f'{", ".join([f"{k}:{v}" for k, v in zip([ast.dump(sub) for sub in first_arg.keys], [ast.dump(sub) for sub in first_arg.values])])}'
                iter_name = f"{{{kw_repr}}}"
            else:  # No obvious name
                if self.verbose:
                    print(type(first_arg))
                    iter_name = f"{ast.dump(first_arg)}"
                else:
                    iter_name = "iterable"
            self.result[node.func.id][node.lineno] = iter_name
        self.generic_visit(node)  # visit the children


def get_first_arg_name(
    func_name: str, *, verbose=False, max_num_intermediate_unnamed_elements=-1  # recurse = -1
) -> Optional[str]:
    """ """
    import inspect
    import textwrap
    import ast

    caller_frame = inspect.currentframe().f_back.f_back
    caller_src_code_lines = inspect.getsourcelines(caller_frame)
    fai = FirstArgIdentifier(
        func_name,
        verbose=verbose,
        max_num_intermediate_unnamed_elements=max_num_intermediate_unnamed_elements,
    )
    fai.visit(ast.parse(textwrap.dedent("".join(caller_src_code_lines[0]))))
    if func_name in fai.result:
        offset = 0
        if caller_src_code_lines[1]:
            offset = caller_src_code_lines[1] - 1
        idx = caller_frame.f_lineno - offset
        if idx in fai.result[func_name]:
            return fai.result[func_name][idx]
        elif verbose:
            print(
                f'Unexpected line number: {idx}, probably a wrong alias "{func_name}" was supplied, found {fai.result[func_name]}, in {inspect.getsourcefile(caller_frame)}'
            )
    elif verbose:
        print(f"{func_name} was not found in {fai.result}")
    return None


def get_first_arg_name_recurse() -> Optional[str]:
    """
    unpack chained generators to base iterator name

    :return:
    :rtype:
    """
    pass  # TODO: For e.g. description in progress_bar(range(_name_))
    raise NotImplementedError


def cprint(v: Any, writer: callable = print, deliminator: str = ":") -> None:
    if isinstance(v, str) and v.strip() == "":
        v = '""'
    writer(f"{get_first_arg_name('cprint')}{deliminator}", v)


if __name__ == "__main__":

    def siajd():
        s = ""
        cprint(s)
        cprint("")
        ass = "    "
        cprint(ass)
        cprint("  ")

    def ausdh() -> None:
        """
        :rtype: None
        """
        import inspect
        import textwrap
        import ast
        from warg import FirstArgIdentifier

        def some_func(a):
            """ """
            caller_frame = inspect.currentframe().f_back
            # caller_src_code_snippet = inspect.getsource(caller_frame) # Only gets scope
            caller_src_code_lines = inspect.getsourcelines(caller_frame)
            caller_src_code_valid = textwrap.dedent(
                "".join(caller_src_code_lines[0])
            )  # TODO: maybe there is a nicer way?
            call_nodes = ast.parse(
                caller_src_code_valid
            )  # parse code to get nodes of abstract syntax tree of the call
            fai = FirstArgIdentifier("some_func")
            fai.visit(call_nodes)
            snippet_offset = caller_src_code_lines[1] - 1
            desc = fai.result["some_func"][caller_frame.f_lineno - snippet_offset]
            print(desc)

        this_name_is_in_another_frame = 5
        this_func_is_in_another_frame = lambda x: x
        this_generator_is_in_another_frame = range(24)

        class this_class_is_in_another_frame:
            pass

        some_func(this_name_is_in_another_frame)
        some_func(this_func_is_in_another_frame)
        some_func(this_generator_is_in_another_frame)
        # some_func([this_func_is_in_another_frame, this_func_is_in_another_frame, this_generator_is_in_another_frame,this_class_is_in_another_frame])
        some_func([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    def ausdh2() -> None:
        """
        :rtype: None
        """
        import inspect
        import textwrap
        import ast
        from warg import FirstArgIdentifier

        def some_func(a):
            """ """
            caller_frame = inspect.currentframe().f_back
            # caller_src_code_snippet = inspect.getsource(caller_frame) # Only gets scope
            caller_src_code_lines = inspect.getsourcelines(caller_frame)
            caller_src_code_valid = textwrap.dedent(
                "".join(caller_src_code_lines[0])
            )  # TODO: maybe there is a nicer way?
            call_nodes = ast.parse(
                caller_src_code_valid
            )  # parse code to get nodes of abstract syntax tree of the call
            fai = FirstArgIdentifier("some_func")
            fai.visit(call_nodes)
            snippet_offset = caller_src_code_lines[1] - 1
            desc = fai.result["some_func"][caller_frame.f_lineno - snippet_offset]
            print(desc)

        some_func({1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7, 8: 8, 9: 9, 10: 10})

    def ausdh3() -> None:
        """
        :rtype: None
        """
        from typing import Any

        def some_func(a: Any) -> None:
            """ """
            print(get_first_arg_name("some_func", verbose=True))

        some_func(print(2, sep="-"))

    def ausd2h3() -> None:
        """
        :rtype: None
        """
        from typing import Any
        import warg

        def some_func(a: Any) -> None:
            """ """
            print(get_first_arg_name("some_func", verbose=True))

        some_func(warg.identity(2))

    def ausd2h3213() -> None:
        """
        :rtype: None
        """
        from typing import Any

        class Ac:
            class Bc:
                @staticmethod
                def c(d):
                    """ """
                    pass

        def some_func(a: Any) -> None:
            """ """
            print(get_first_arg_name("some_func", verbose=True))

        some_func(Ac.Bc.c(2))

    # ausdh()
    # ausdh2()
    # ausdh3()
    # ausd2h3()
    # ausd2h3213()
    siajd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"
__doc__ = r"""
          Plugin

           Created on 13/06/2020
           """

from typing import Sequence, Tuple

__all__ = ["split"]


def split(seq: Sequence) -> Tuple[Sequence, Sequence]:
    m = len(seq) // 2
    return seq[:m], seq[m:]


if __name__ == "__main__":
    print((split(list(range(11)))))
    print((split(list(range(10)))))
    print((split(list(range(9)))))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "Christian Heider Nielsen"
__doc__ = ""

__all__ = [
    "IterDictItemsMixin",
    "IterDictKeysMixin",
    "IterDictValuesMixin",
]


class IterDictItemsMixin:
    """
    Mixin class for iterating kw pairs in a class instance __dict__"""

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value


class IterDictKeysMixin:
    """
    Mixin class for iterating only the keys of a class instance __dict__"""

    def __iter__(self):
        for attr in self.__dict__.keys():
            yield attr


class IterDictValuesMixin:
    """
    Mixin class for iterating only the values of a class instance __dict__"""

    def __iter__(self):
        for value in self.__dict__.values():
            yield value


if __name__ == "__main__":

    def asdij() -> None:
        """
        :rtype: None
        """

        class IASD(IterDictValuesMixin):
            pass

        a = IASD()
        a.b = 1
        a.c = 2
        a.d = 3

        for ca in a:
            print(ca)

    asdij()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import argparse
from collections import namedtuple
from pathlib import Path, PosixPath
from typing import Mapping, Tuple
from warnings import warn

from warg.data_structures.named_ordered_dictionary import NOD

__author__ = "Christian Heider Nielsen"

__all__ = [
    "to_lower_properties",
    "get_upper_case_vars_or_protected_of",
    "config_to_mapping",
    "add_bool_arg",
    "check_for_duplicates_in_args",
    "UpperAttrMetaclass",
]


class UpperAttrMetaclass(type):
    """
    Upper case all attributes if not __private"""

    def __new__(mcs, clsname, bases, dct: dict):

        uppercase_attr = {}
        for name, val in dct.items():
            if not name.startswith("__"):
                uppercase_attr[name.upper()] = val
            else:
                uppercase_attr[name] = val

        return super(UpperAttrMetaclass, mcs).__new__(mcs, clsname, bases, uppercase_attr)


class ConfigObject(object):
    """
    Config object"""

    pass


def to_lower_properties(C_dict: Mapping):
    """

    :param C_dict:
    :type C_dict:
    :return:
    :rtype:"""
    if not isinstance(C_dict, dict):
        C_dict = config_to_mapping(C_dict)

    a = ConfigObject()

    for (k, v) in C_dict.items():
        assert isinstance(k, str)
        lowered = k.lower()
        if isinstance(v, (PosixPath, Path)):
            setattr(a, lowered, str(v))
        else:
            setattr(a, lowered, v)

    return a


def lower_dict(mapping: Mapping) -> Mapping:
    """

    :param mapping:
    :type mapping:
    :return:
    :rtype:"""
    cop = {}
    for (k, v) in mapping.items():
        assert isinstance(k, str)
        cop[k.lower()] = v

    return cop


def upper_dict(mapping: Mapping) -> Mapping:
    """

    :param mapping:
    :type mapping:
    :return:
    :rtype:"""
    cop = {}
    for (k, v) in mapping.items():
        assert isinstance(k, str)
        cop[k.upper()] = v

    return cop


def get_upper_case_vars_or_protected_of(module: object, lower_keys: bool = True) -> Mapping:
    """

    :param module:
    :type module:
    :param lower_keys:
    :type lower_keys:
    :return:
    :rtype:"""
    v = vars(module)
    check_for_duplicates_in_args(**v)
    if v:
        if lower_keys:
            return {
                key.lower(): value
                for key, value in module.__dict__.items()
                if (key.isupper() or (key.startswith("_")) and not key.endswith("_"))
            }
        return {
            key: value
            for key, value in module.__dict__.items()
            if (key.isupper() or (key.startswith("_")) and not key.endswith("_"))
        }
    return {}


def config_to_mapping(C: object, only_upper_case: bool = True) -> NOD:
    """

    :param C:
    :type C:
    :param only_upper_case:
    :type only_upper_case:
    :return:
    :rtype:"""
    if only_upper_case:
        return NOD(get_upper_case_vars_or_protected_of(C))
    else:
        return NOD(vars(C))


def add_bool_arg(
    parser: argparse.ArgumentParser,
    name: str,
    *,
    dest: str = None,
    converse: str = None,
    default: bool = False,
    **kwargs,
):
    """

    :param parser:
    :type parser:
    :param name:
    :type name:
    :param dest:
    :type dest:
    :param converse:
    :type converse:
    :param default:
    :type default:
    :param kwargs:
    :type kwargs:"""
    if not dest:
        dest = name

    group = parser.add_mutually_exclusive_group(required=False)

    group.add_argument(
        f"--{name.upper()}",
        f"-{name.lower()}",
        dest=dest,
        action="store_true",
        **kwargs,
    )
    if converse:
        group.add_argument(
            f"--{converse.upper()}",
            f"-{converse.lower()}",
            dest=dest,
            action="store_false",
            **kwargs,
        )
    else:
        group.add_argument(
            f"--NO-{name.upper()}",
            f"-no-{name.lower()}",
            dest=dest,
            action="store_false",
            **kwargs,
        )
    parser.set_defaults(**{dest: default})


def check_for_duplicates_in_args(**kwargs) -> None:
    """

    :param kwargs:
    :type kwargs:"""
    for key, value in kwargs.items():

        occur = 0

        if kwargs.get(key) is not None:
            occur += 1
        else:
            pass

        if key.isupper():
            k_lowered = f"_{key.lower()}"
            if kwargs.get(k_lowered) is not None:
                occur += 1
            else:
                pass
        else:
            k_lowered = f'{key.lstrip("_").upper()}'
            if kwargs.get(k_lowered) is not None:
                occur += 1
            else:
                pass

        if occur > 1:
            warn(f"Config contains hiding duplicates of {key} and {k_lowered}, {occur} times")


def str_to_bool(s: str, truthies: Tuple[str, ...] = ("true", "1")) -> bool:
    """


    :param truthies:
    :param s:
    :return:"""
    return s.lower() in truthies


str2bool = str_to_bool

if __name__ == "__main__":
    c = namedtuple("C", ("a", "b"))

    def add2(a, b):
        """

        :param a:
        :type a:
        :param b:
        :type b:
        :return:
        :rtype:"""
        return a + b

    wq = add2(2, 4)
    print(wq)

    wc = add2(*c(4, 3))
    print(wc)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "heider"
__doc__ = r"""

           Created on 01/02/2022
           """

from pathlib import Path

with open(Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .color_conversion import *
from .css_colors import *
from .label_colors import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "Christian Heider Nielsen"
__doc__ = r"""
          This file is not for general use. Ode to python

           Created on 13/06/2020
           """

__all__ = []

if __name__ == "__main__":
    # from __future__ import braces # TODO: Try disable comment

    # noinspection PyUnresolvedReferences
    import __hello__

    print()
    import this

    print()
    # noinspection PyUnresolvedReferences
    import antigravity

    print()
    print(this.i, this.d, this.c, this.s)
    print()

    # noinspection PyUnresolvedReferences
    import that  # pip install that



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 21/12/2019
           """

from collections import defaultdict
from typing import Dict, Optional

__all__ = ["AutoDict", "recursive_default_dict_print", "sanitise_auto_dict"]


def AutoDict() -> defaultdict:
    """

    :return:
    :rtype:"""
    return defaultdict(autodict)


def recursive_default_dict_print(d: Dict, depth: int = 1) -> None:
    """

    :param d:
    :type d:
    :param depth:
    :type depth:"""
    for k, v in d.items():
        print("-" * depth, k)
        if type(v) is defaultdict:
            recursive_default_dict_print(v, depth + 1)
        else:
            print("-" * (depth + 1), v)


def sanitise_auto_dict(d: Dict) -> Optional[Dict]:
    """

    :param d:
    :type d:
    :return:
    :rtype:"""
    if isinstance(d, defaultdict):
        if len(d.keys()) == 0:
            return None

    out_dict = dict()
    for k, v in d.items():
        if isinstance(v, defaultdict):
            sanitised = sanitise_auto_dict(v)
            if sanitised is None:
                continue
            out_dict[k] = sanitised
        else:
            out_dict[k] = v
    if len(out_dict) > 0:
        return out_dict
    return None


autodict = AutoDict
AD = AutoDict

if __name__ == "__main__":
    ad = AutoDict()

    a = ad["b"]["b"]["b"]["b"]
    ad["b"]["b"]["c"]["b"]["c"] = {}
    ad["c"] = 1
    ad["cd"]["v"] = 1
    ad["qc"]["d"] = []
    ad["cf"]["b6"] = None
    ad["cf"]["1"] = None

    print(ad)

    recursive_default_dict_print(ad)

    print(sanitise_auto_dict(ad))



import copy

DictProxyType = type(object.__dict__)

__all__ = ["make_hash"]


def make_hash(o):
    """
    Makes a hash from a dictionary, list, tuple or set to any level, that
    contains only other hashable types (including any lists, tuples, sets, and
    dictionaries). In the case where other kinds of objects (like classes) need
    to be hashed, pass in a collection of object attributes that are pertinent.
    For example, a class can be hashed in this fashion:

      make_hash([cls.__dict__, cls.__name__])

    A function can be hashed like so:

      make_hash([fn.__dict__, fn.__code__])"""

    if isinstance(o, DictProxyType):
        o2 = {}
        for k, v in o.items():
            if not k.startswith("__"):
                o2[k] = v
        o = o2

    if isinstance(o, (set, tuple, list)):
        return hash(tuple([make_hash(e) for e in o]))
    if not isinstance(o, dict):
        return hash(o)

    new_o = copy.deepcopy(o)
    for k, v in new_o.items():
        new_o[k] = make_hash(v)

    return hash(tuple(frozenset(sorted(new_o.items()))))


if __name__ == "__main__":
    print(hash(1))
    print(make_hash(1))
    print(make_hash(1))
    print(make_hash({1}))
    print(make_hash([1]))
    print(make_hash({1}))
    print(make_hash({1, 2}))
    print(make_hash([1, 2]))
    print(make_hash((1, 2)))
    print(make_hash({4}))
    print(make_hash("1"))
    print(make_hash({"2": 2}))
    print(make_hash({"2": 3}))
    print(make_hash({"3": 2}))
    print(make_hash({"3": 3}))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"
__doc__ = r"""
          Plugin

           Created on 13/06/2020
           """

from typing import Tuple

import pkg_resources

__all__ = ["get_plugins", "get_static_plugins", "get_dynamic_plugins"]


def get_plugins(package_name: str) -> Tuple:
    """Returns a list specifying all known plugins.

    This includes both first-party, statically bundled plugins and
    dynamic plugins.

    Returns:
      The list of default first-party plugins.
    """
    return (*get_static_plugins(package_name), *get_dynamic_plugins(package_name))


def get_static_plugins(package_name: str) -> Tuple:
    """Returns a list specifying  default first-party plugins.

      DECLARE GLOBAL (PACKAGE_NAME)_PLUGINS tuple with entries

    Returns:
      The list of default first-party plugins.

    """
    v = f"{package_name.upper()}_PLUGINS"
    if v in globals():
        return globals()[v][:]
    return ()


def get_dynamic_plugins(package_name: str) -> Tuple:
    """Returns a list specifying  dynamically loaded plugins.

    Returns:
      The list of dynamic plugins.

    [1]: https://packaging.python.org/specifications/entry-points/
    """

    # .load() method to import and load that entry point (module or object).
    # from importlib import metadata # new method!
    # return [      entry_point.load()      for entry_point in metadata.entry_points()[f'{package_name}_plugins']      ]
    return (entry_point.load() for entry_point in pkg_resources.iter_entry_points(f"{package_name}_plugins"))


if __name__ == "__main__":
    print(get_plugins("warg"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11-12-2020
           """

if __name__ == "__main__":

    async def a() -> None:
        """
        :rtype: None
        """
        import addition_config_usage

        await addition_config_usage.b()
        import config2

        print(config2.ANOTHER_CONSTANT)

    import asyncio

    asyncio.run(a())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28/06/2020
           """

__all__ = ["is_module_available", "import_warning", "reimported_warning"]

from importlib.util import find_spec
from warnings import warn


def is_module_available(module: str) -> bool:
    """**Return True if module is available.**

    Parameters
    ----------
    module: str
          Name of the module to be checked.

    Returns
    -------
    bool
          True if installed.
    """
    return find_spec(module) is not None


def import_warning(module_name: str) -> None:
    """
    Inform the user that module has been imported,
    useful when repeated imports are heavy in the contexts of multiprocessing.
    Lets the user identify which file is reporting heavy loading and restructure code to avoid repeated importing

    :param module_name:
    :return:
    """
    from sys import modules

    if module_name in modules:
        warn(f"You have imported {module_name}")


def reimported_warning(module_name: str) -> None:
    """
    Just an idea

    :return:
    """
    raise NotImplemented
    # TODO: touch .lock file to system for module_name for a multiprocess warning if already exists,
    # delete it again once process is done?
    # context_wrapper maybe useful


if __name__ == "__main__":

    def main() -> None:
        """
        :rtype: None
        """
        mod = "matplotlib"
        import_warning(mod)
        from matplotlib import pyplot

        import_warning(mod)
        pyplot.figure()

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 13-01-2021
           """

__all__ = ["lcm"]

try:
    from math import lcm
except:
    from math import gcd
    from warg import Number

    def lcm(a: Number, b: Number) -> Number:
        """Least common multiple"""
        return abs(a * b) / gcd(a, b) if a and b else 0



import math

from warg.typing_extension import Number

__all__ = [
    "ceil_odd",
    "ceil_even",
    "floor_even",
    "floor_odd",
    "next_even",
    "next_odd",
    "prev_even",
    "prev_odd",
]


def ceil_even(v: Number) -> Number:
    return math.ceil(v / 2.0) * 2


def floor_even(v: Number) -> Number:
    return math.floor(v / 2.0) * 2


def ceil_odd(v: Number) -> Number:
    return math.ceil(v) // 2 * 2 + 1


def floor_odd(v: Number) -> Number:
    v = math.floor(v)
    return v + v % 2 - 1


def next_even(v: Number) -> Number:
    return floor_even(v) + 2


def next_odd(v: Number) -> Number:
    return floor_odd(v) + 2


def prev_even(v: Number) -> Number:
    return ceil_even(v) - 2


def prev_odd(v: Number) -> Number:
    return ceil_odd(v) - 2


if __name__ == "__main__":
    from warg.ast_ops import cprint
    import numpy

    for a in numpy.arange(-1.0, 10.0, 0.5):
        cprint(a)
        for f in (ceil_odd, ceil_even, floor_even, floor_odd, next_even, next_odd, prev_even, prev_odd):
            print(f.__name__, f(a))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import sys

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
          get os platform utilities 

           Created on 13/06/2020
           """

__all__ = ["is_nix", "is_windows", "is_mac", "is_linux", "get_platform"]

CUR_OS = sys.platform
IS_WIN = any(CUR_OS.startswith(i) for i in ["win32", "cygwin"])
IS_LINUX = CUR_OS.startswith("linux")
IS_MAC = CUR_OS.startswith("darwin")
IS_NIX = IS_LINUX or IS_MAC or CUR_OS.startswith("aix")


def is_windows() -> bool:
    """

    :return:
    """
    return IS_WIN


def is_linux() -> bool:
    """

    :return:
    """
    return IS_LINUX


def is_nix() -> bool:
    """

    :return:
    """
    return IS_NIX


def is_mac() -> bool:
    """

    :return:
    """
    return IS_MAC


def get_platform() -> str:
    """

    :return:
    """
    return CUR_OS



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 18/07/2020
           """

__all__ = [
    "Number",
    "Reals",
    "Numbers",
    "Single",
    "Double",
    "Triple",
    "Quad",
    "Quint",
    "SingleNumber",
    "DoubleNumber",
    "TripleNumber",
    "QuadNumber",
    "QuintNumber",
    "StrictNumbers",
]

import numbers
from typing import Any, Sequence, Tuple, Union

Number = Union[int, float]
Reals = Sequence[Union[numbers.Real, "Reals"]]
Numbers = Sequence[Union[Number, "Numbers"]]
StrictNumbers = Union[Union[Sequence[int], "StrictNumbers"], Union[Sequence[float], "StrictNumbers"]]

Single = Tuple[Any]
Double = Tuple[Any, Any]
Triple = Tuple[Any, Any, Any]
Quad = Tuple[Any, Any, Any, Any]
Quint = Tuple[Any, Any, Any, Any, Any]

SingleNumber = Tuple[Number]
DoubleNumber = Tuple[Number, Number]
TripleNumber = Tuple[Number, Number, Number]
QuadNumber = Tuple[Number, Number, Number, Number]
QuintNumber = Tuple[Number, Number, Number, Number, Number]

if __name__ == "__main__":

    def stest() -> None:
        """
        :rtype: None
        """
        assert (
            isinstance(1, Number.__args__)
            and isinstance(1.1, Number.__args__)
            and not isinstance(complex(1, 1), Number.__args__)
        )
        # assert isinstance(list(range(2)), Numbers.__args__) and isinstance((float(i) for i in range(2)), Numbers.__args__) and isinstance((1,2.0), Numbers.__args__)
        # assert isinstance(list(range(2)), StrictNumbers.__args__) and isinstance((float(i) for i in range(2)), StrictNumbers.__args__) and not isinstance((1,2.0), StrictNumbers.__args__)
        assert (
            isinstance(1, numbers.Real)
            and isinstance(1.1, numbers.Real)
            and not isinstance(complex(2, 2), numbers.Real)
        )
        # assert isinstance(list(range(2)), Reals.__args__) and isinstance((float(i) for i in range(2)), Reals.__args__) and isinstance((1,2.0), Reals.__args__)

    stest()
